import { generateGUID, pushUsedGUID } from "./guid-pool";
import { getObservable } from "./index";
import { RxPostmessengerRequest } from "./request";
import { includes, isString } from "./vendor/lodash-es";
/**
 * @class RxPostmessenger
 */
var Messenger = /** @class */ (function () {
    /**
     * @param {Window} otherWindow - The window object to exchange messages with.
     * @param {string} origin - The remote url to accept incoming messages from.
     */
    function Messenger(otherWindow, origin) {
        var _this = this;
        this.otherWindow = otherWindow;
        this.origin = origin;
        // Initialize read-only properties
        this.inboundMessages$ = getObservable()
            .fromEvent(window, "message")
            .filter(function (message) { return _this.isValidMessage(message); })
            .pluck("data");
        this.requests$ = this.messagesOfType("request");
        this.responses$ = this.messagesOfType("response");
        this.notifications$ = this.messagesOfType("notification");
        // Other initializers
        this.syncGUIIDs();
    }
    // ------------------------------------------------------------------------------------------------
    // API
    // ------------------------------------------------------------------------------------------------
    /**
     * @param {string} channel
     * @param {*} payload
     *
     * @return {Observable<object>}
     * @public
     */
    Messenger.prototype.request = function (channel, payload) {
        var id = generateGUID();
        var responseObservable = this.createResponseObservable(id);
        this.postMessage({
            channel: channel,
            id: id,
            payload: payload,
            type: "request",
        });
        return responseObservable;
    };
    /**
     * Sends a notification through given channel to the remote window, carrying given payload.
     *
     * @param {string} channel
     * @param {*} payload
     * @return {Messenger}
     * @public
     */
    Messenger.prototype.notify = function (channel, payload) {
        this.postMessage({
            channel: channel,
            id: generateGUID(),
            payload: payload,
            type: "notification",
        });
        return this;
    };
    /**
     * Returns an Observable that emits the subset of inbound notification messages where their channel equals given
     * channel name.
     *
     * @param {string} channel
     * @return {Observable<object>}
     * @public
     */
    Messenger.prototype.requests = function (channel) {
        var _this = this;
        return this.requests$
            .filter(function (request) { return request.channel === channel; })
            .map(function (request) { return new RxPostmessengerRequest(request.id, request.channel, request.payload, function (payload) { return _this.respond(request.id, channel, payload); }); });
    };
    /**
     * Returns an Observable that emits the subset of inbound notification messages where their channel equals given
     * channel name.
     *
     * @param {string} channel
     * @return {Observable<*>}
     * @public
     */
    Messenger.prototype.notifications = function (channel) {
        return this.notifications$
            .filter(function (notification) { return notification.channel === channel; })
            .pluck("payload");
    };
    // ------------------------------------------------------------------------------------------------
    // Private runtime
    // ------------------------------------------------------------------------------------------------
    /**
     * Sends a response through given channel to the remote window, carrying given payload.
     *
     * @param {string} requestId
     * @param {*} payload
     * @return {Messenger}
     * @private
     */
    Messenger.prototype.respond = function (requestId, channel, payload) {
        this.postMessage({
            channel: channel,
            id: generateGUID(),
            payload: payload,
            requestId: requestId,
            type: "response",
        });
        return this;
    };
    /**
     * Creates an Observable that completes after a single emission of the MessageEvent response for request of given
     * requestId. If no matching response is ever received, the Observable never completes.
     *
     * @param {string} requestId
     * @return {Observable<object>}
     * @private
     */
    Messenger.prototype.createResponseObservable = function (requestId) {
        return this.responses$
            .filter(function (response) { return response.requestId === requestId; })
            .pluck("payload")
            .take(1);
    };
    /**
     * Validates the identity of the message's sender and the format of the message's data.
     *
     * Chechs whether the origin location matches any allowed origins.
     * Separate assertion of the origin allows for cross-domain navigation
     * within this.frame, and still treating inbound messages from the
     * frame as being equal.
     *
     * Checks whether the source Window object equals the remoteWindow object.
     * This check allows for implementation of multiple iframes that share the
     * same origin, and still being able to distinguish between messages from
     * such frames.
     *
     * @param {MessageEvent} message
     * @return {boolean}
     * @private
     */
    Messenger.prototype.isValidMessage = function (message) {
        return message instanceof MessageEvent
            && message.origin === this.origin
            && message.source === this.otherWindow
            && this.isWellFormedMessage(message.data);
    };
    /**
     * Tests whether the data sent through postMessage is a well-formed message
     * object. This serves as runtime data format validation. If messages do not
     * comply to the AnyMessage compound interface, the entire event is ignored.
     *
     * @param {AnyMessage} message
     * @return {boolean}
     */
    Messenger.prototype.isWellFormedMessage = function (message) {
        return isString(message.id)
            && includes(["request", "response", "notification"], message.type)
            && isString(message.channel)
            && (message.type !== "response" || isString(message.requestId));
    };
    /**
     * Performs a postMessage call with given data to this.otherWindow, provided that its
     * location origin matches this.origin.
     *
     * @param {IMessageObject} data
     * @private
     */
    Messenger.prototype.postMessage = function (data) {
        this.otherWindow.postMessage(data, this.origin);
    };
    // ------------------------------------------------------------------------------------------------
    // Init helpers
    // ------------------------------------------------------------------------------------------------
    /**
     * Starts a GUID sync that intercepts incoming messages from other windows running this package,
     * and pushes their ID values into a used IDs array. This way, every message should have a unique ID.
     *
     * @return {Messenger}
     * @private
     */
    Messenger.prototype.syncGUIIDs = function () {
        this.inboundMessages$.subscribe(function (_a) {
            var id = _a.id;
            return pushUsedGUID(id);
        });
        return this;
    };
    /**
     * Returns an Observable emitting all inbound messages` of given type.
     *
     * Returns an Observable emitting a subset of MessageEvent objects emitted
     * by this.inboundMessages$, passing through all MessageEvent objects whose
     * data.type property matches given type.
     *
     * @param {string} type
     * @return {Observable<object>}
     * @private
     */
    Messenger.prototype.messagesOfType = function (type) {
        return this.inboundMessages$
            .filter(function (message) { return message.type === type; });
    };
    return Messenger;
}());
export { Messenger };
//# sourceMappingURL=messenger.js.map